#include <gtk/gtk.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <gtkgl/gtkglarea.h>
#include "dnd_glarea.h"
#include "trackball.h"
#include "dnd_graphics.h"
#include "jitter.h"
#include "vector3.h"

#define MAX_LIGHTS 7

enum light_types
{
  OGL_DIRECTIONAL_LIGHT,
  OGL_POINT_LIGHT,
  OGL_SPOT_LIGHT
};


/* ---- */

typedef struct _mesh_info MeshInfo;

struct _mesh_info
{
  double mouse_x;
  double mouse_y;
  double aspect_window_x;
  double aspect_window_y;
  double scale;  
  double fovy;   /* angle in gulPerspective (GLdouble fovy, ...) */
  double slab;
  GLfloat trans_x, trans_y;
  GLfloat quat[4];
  double window;
  GLfloat xform[4][4];
};

static MeshInfo mesh_info;


/* --- */

static void ogl_render_lights (void);
static void ogl_render_lists (void);

/* --- */

static GSList *label_display_slist = NULL;
static GSList *alpha_display_slist = NULL;
static GSList *display_slist = NULL;
static int latest_display_list;
static GLfloat output_width;
static GLfloat output_height;
/*
static float black_colour[3] = {0.0, 0.0, 0.0 };
static float background_colour[3] = {0.0, 0.0, 0.0 };
*/
static int accum = 0;
static jitter_point *jitter_points;

static GLuint pick_buffer[512];
static int    pick_atom = 0;

static GLuint fontbase;
static gboolean is_label_show = FALSE;

static gint
init (GtkWidget *widget)
{
  GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
  GLfloat mat_shininess[] = { 50.0 };
  GLfloat light_position[] = { 1.0, 1.0, 1.0, 0.0 };
  GdkFont *font;
  char *fontname = "-adobe-helvetica-medium-r-normal--*-120-*-*-*-*-*-*";


  if (gtk_gl_area_make_current (GTK_GL_AREA (widget)))
    {
      glClearColor (0.0, 0.0, 0.0, 0.0);
      glShadeModel (GL_SMOOTH);

      glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
      glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
      glLightfv(GL_LIGHT0, GL_POSITION, light_position);

      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
      glEnable(GL_DEPTH_TEST);

      /*  generate font display lists */
      font = gdk_font_load (fontname);
      
      if (!font)
	{
	  g_print("Can't load font '%s'\n", fontname);
	  gtk_exit(1);
	}

      gdk_gl_use_gdk_font (font, 0, 128, fontbase);
      /*      fontheight = font->ascent + font->descent; */
      gdk_font_unref (font);
      
      /* -- initiate mesh_info -- */
      mesh_info.fovy = 45.0;
      mesh_info.slab = 40;
      mesh_info.trans_x = 0;
      mesh_info.trans_y = 0;
      mesh_info.scale = 1.;
      mesh_info.aspect_window_x = 20;
      mesh_info.aspect_window_y = 20;
      
      project_option = ORTHO_MODE;
      mouse_option = ROTATE_MODE;
      
    }
  
  return TRUE;
}


static void 
init_gl (GraphicsInfo *g_info)
{
  GList *list;
  LightInfo *light_info;
  GLfloat  light_color[4];
  
  /* remove back faces */
  glDisable (GL_CULL_FACE);
  glEnable (GL_DEPTH_TEST);

  /* speedups */
  glEnable(GL_DITHER);
  glShadeModel(GL_SMOOTH);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);

  for (list = g_info->light_list; list; list = list->next)
    {
      light_info = list->data;
      
      glLightfv (light_info->light_enum, GL_POSITION,
		 light_info->light_pos);
      light_color[0] = light_info->light_color.r;
      light_color[1] = light_info->light_color.g;
      light_color[2] = light_info->light_color.b;
      light_color[3] = light_info->light_color.a;
      		 
      glLightfv (light_info->light_enum, GL_DIFFUSE, 
		 light_color);
      glEnable (light_info->light_enum);
    }
  glEnable(GL_LIGHTING);
  
  glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);
  glMaterialf(GL_FRONT, GL_SHININESS, 20.0);
  
}


void
start_display_list (gboolean is_label)
{
  GLint *num;
    
  num = g_new0 (GLint, 1);
  *num = glGenLists (1);
  
  glNewList (*num, GL_COMPILE);
    
  if (is_label)
    {
      label_display_slist = g_slist_append (label_display_slist, num);
    }
  /*  else if (current_state->transparency != 0.0)
    {
      alpha_display_slist = g_slist_append (alpha_display_slist, num);
      }*/
  else
    {
      display_slist = g_slist_append (display_slist, num);
    }
}


void
delete_display_lists ()
{
  GLint *num;
  GSList *slist;
  
  for (slist = label_display_slist; slist; slist = slist->next)
    {
      num = slist->data;
      glDeleteLists (*num, 1);
    }
  g_slist_free (label_display_slist);
  label_display_slist = NULL;
  
  for (slist = display_slist; slist; slist = slist->next)
    {
      num = slist->data;
      glDeleteLists (*num, 1);
    }
  g_slist_free (display_slist);
  display_slist = NULL;
}


static void
ogl_render_init (int render_mode)
{

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
 
  if (render_mode == GL_SELECT && mouse_option == PICK_MODE)
    {
      GLint viewport[4];
      
      glGetIntegerv (GL_VIEWPORT, viewport);
      gluPickMatrix (mesh_info.mouse_x,
		     viewport[3] - mesh_info.mouse_y,
		     5.0, 5.0, viewport);
    }
  
  switch (project_option)
    {
    case PERSPECTIVE_MODE:
      {
	gluPerspective (mesh_info.fovy, 1.0, 1.0, 100.0);
	break;
      }
    case ORTHO_MODE:
      {
	
	glOrtho (-20 * mesh_info.scale,
		 20  * mesh_info.scale,
		 -20 * mesh_info.scale,
		 20  * mesh_info.scale,
		 0, 80);

	break;
      }
    }
    
  glMatrixMode(GL_MODELVIEW);

  /* [FIXME] 
  glLoadIdentity();
  gluLookAt (0.0, 0.0, 40,
	     0.0, 0.0, 0.0,
	     0.0, 1.0, 0.0);
  */
  /* clear background */
  glClearColor(0,0,0,1);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glDepthFunc(GL_LESS);
  glEnable(GL_DEPTH_TEST);
}

static gint
glarea_configure (GtkWidget * widget, GdkEventConfigure * event)
{
  /* OpenGL calls can be done only if begingl returns true */
  if (gtk_gl_area_make_current(GTK_GL_AREA (widget)))
    {
      output_width = widget->allocation.width;
      output_height = widget->allocation.width;
      glViewport (0, 0, output_width, output_height);
    }
  return TRUE;
}


static void
pick_start (void)
{
  glSelectBuffer (512, pick_buffer);
  glRenderMode (GL_SELECT);
  
  glInitNames ();
  glPushName (0);
  
}


static void
pick_end (DNDData *data)
{
  int hits = glRenderMode (GL_RENDER);
  /*  
  g_print ("hits = %d \n", hits);
  */
  if (hits > 1)
    {
      is_label_show = TRUE;
      pick_atom = pick_buffer[(hits-1)*4 + 3];
      /*
      if (g_slist_find (label_slist, &n) == NULL)
	{
	  num = g_new0 (int, 1);
	  *num  = n;
	  label_slist = g_slist_prepend (label_slist, num);
	}
      */
    }
  
}


/*------------------------------------------------------------*/
static gint
glarea_button_press (GtkWidget * widget, GdkEventButton * event)
{
  gint button;
  GtkWidget *menu;
  GtkWidget *app_win;
    
  button = event->button;

  if (button == 3)
    {
      app_win = (GtkWidget *) gtk_object_get_data (GTK_OBJECT (widget),
						   "widget");
      menu = (GtkWidget *)gtk_object_get_data(GTK_OBJECT(app_win),
					      "popup_menu");
      if (menu)
	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, 3, event->time);
      return TRUE;
    }
  else if (button == 1)
    {
      mesh_info.mouse_x = event->x;
      mesh_info.mouse_y = event->y;
    }
  
  return TRUE;
}


static gint
glarea_button_release (GtkWidget *widget, GdkEventButton *event)
{
  DNDData   *data;
  GraphicsInfo *g_info;
  gint       button;
  
  data = (DNDData *) gtk_object_get_data (GTK_OBJECT (widget),
					  "dnd_data");
  g_info = data->graphics_info;
  
  button = event->button;
  
  if (button == 1 && mouse_option == PICK_MODE)
    {
      is_label_show = FALSE;
      
      if (g_info->render_mode == GL_RENDER)
	{
	  g_info->render_mode = GL_SELECT;
	  
	  dnd_display_pick (data);
	}
      
      gtk_widget_draw (widget, NULL);
            
      if (is_label_show)
	{
	  Pos *pos = (Pos *) g_list_nth_data (data->mol->pos_list, pick_atom);
	  Atom *atom = (Atom *) g_list_nth_data (data->mol->atom_list, pick_atom);
	  dnd_display_label (pos, atom->name, fontbase);
	  
	  /* Tricks to show label on the OpenGL window */
	  g_info->render_mode = GL_RENDER;
	  gtk_widget_draw (widget, NULL);
	  g_info->render_mode = GL_SELECT;
	}
      
    }

  return TRUE;
}



/*------------------------------------------------------------*/
static gint
glarea_motion_notify (GtkWidget * widget, GdkEventMotion * event)
{
  GLfloat spin_quat[4];
  int    x, y;
  GdkModifierType state;
  GdkRectangle area;
  
  if (event->is_hint)
    gdk_window_get_pointer (event->window, &x, &y, &state);
  else
    {
      x = event->x;
      y = event->y;
      state = event->state;
    }
    
  area.x = 0;
  area.y = 0;
  area.width = widget->allocation.width;
  area.height = widget->allocation.height;

  if (state & GDK_BUTTON1_MASK)
    {
      switch (mouse_option)
	{
	case TRANSLATE_MODE: 
	  {
	    mesh_info.trans_x += ((x - mesh_info.mouse_x) / area.width) * 10;
	    mesh_info.trans_y -= ((y - mesh_info.mouse_y) / area.height) * 10;
	    gtk_widget_draw (widget, &area);
	    break;
	  }
	case ROTATE_MODE: 
	  {
	    trackball (spin_quat, 
		       (2.0*mesh_info.mouse_x -       area.width) / area.width,
		       (     area.height - 2.0*mesh_info.mouse_y) / area.height,
		       (           2.0*x -       area.width) / area.width,
		       (     area.height -            2.0*y) / area.height);
	    add_quats(spin_quat, mesh_info.quat, mesh_info.quat);
	    
	    gtk_widget_draw (widget, &area);
	    
	    break;
	  }
	case ZOOM_MODE:
	  {
	    switch (project_option)
	      {
	      case PERSPECTIVE_MODE:
		{
		  mesh_info.fovy += ((y - mesh_info.mouse_y)/area.height) *40;
		  if (mesh_info.fovy < 5)
		    mesh_info.fovy = 5;
		  if (mesh_info.fovy > 120)
		    mesh_info.fovy = 120;
		  break;
		}
	      case ORTHO_MODE:
		{
		  mesh_info.scale += ((y - mesh_info.mouse_y)/area.height)*2;
		  if (mesh_info.scale < 1.0e-2)
		    mesh_info.scale = 1.0e-2;
		  break;
		}
	      }
	    
	    gtk_widget_draw (widget, &area);
	    break;
	  }
	  	  
	  /*    case SLAB_MODE:
		{
		mesh_info.slab *= 1.0 + 0.004 * (mesh_info.mouse_y - y);
		if (mesh_info.slab < 1.0e-4)
		mesh_info.slab = 1.0e-4;
		
		gtk_widget_draw (widget, &area);
		break;
		} */
	default:
	  break;
      
	}
    }
  
  mesh_info.mouse_x = x;
  mesh_info.mouse_y = y;

  return TRUE;

}

static void
window_render (DNDData *data)
{
  int render_mode = data->graphics_info->render_mode;
  
  if (render_mode == GL_SELECT && mouse_option == PICK_MODE)
    pick_start ();
  
  /* -- view -- */
  ogl_render_init (render_mode);

  /* --  draw object -- */
  glLoadIdentity ();
  
  switch (project_option)
    {
    case PERSPECTIVE_MODE:
    case ORTHO_MODE:
      glTranslatef (mesh_info.trans_x, mesh_info.trans_y, -30);
    }
  
  build_rotmatrix (mesh_info.xform, mesh_info.quat);
  glMultMatrixf (&(mesh_info.xform[0][0]));

  ogl_render_lists ();

  if (render_mode == GL_SELECT && mouse_option == PICK_MODE) 
    pick_end (data);
  
}


/* ------------------------------------------------------------ */

static gint
glarea_expose (GtkWidget * widget, GdkEventExpose * event)
{
  GtkGLArea *glarea = GTK_GL_AREA (widget);
  DNDData   *data;
  
  data = (DNDData *) gtk_object_get_data (GTK_OBJECT (widget),
					  "dnd_data");
  if (event->count > 0)
    return FALSE;

  if (gtk_gl_area_make_current (glarea))
    {
      window_render (data);
      
      gtk_gl_area_swapbuffers (glarea);
    }

  return TRUE;
  
}

static gint
glarea_destroy (GtkWidget *widget)
{
  return TRUE;
}


/*------------------------------------------------------------*/
static void
ogl_render_lights (void)
{
  GLfloat LightPos[4] = {-10.0, 10.0, 0.0, 0.0};
  /*  GLfloat light0_color[4] =  { .6, .6, .6, 1.0 }; */
  /*  GLfloat Black[4] = {0.0, 0.0, 0.0, 1.0}; */
  GLfloat White[4] = {1.0, 1.0, 1.0, 1.0};

  glLightfv (GL_LIGHT0, GL_POSITION, LightPos);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, White);
  glEnable (GL_LIGHT0);
  glEnable (GL_LIGHTING);
  /*
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 0);
  */
  
 
  glMaterialfv(GL_FRONT, GL_DIFFUSE, White);
  glMaterialfv(GL_FRONT, GL_SPECULAR, White);
  glMaterialf(GL_FRONT, GL_SHININESS, 20.0);

}


/*------------------------------------------------------------*/
void
ogl_render_lists (void)
{
  int num;
  double aspect_window_x, aspect_window_y;
  double window_scale;
  GSList *slist;
  
  aspect_window_x = mesh_info.aspect_window_x;
  aspect_window_y = mesh_info.aspect_window_y;
  window_scale    = mesh_info.scale;

  if (accum != 0)
    {
      int point;
      GLfloat fraction = 1.0 / (GLfloat) accum;
      float xpix = 2.0 * aspect_window_x * window_scale / (float) output_width;
      float ypix = 2.0 * aspect_window_y * window_scale / (float) output_height;
      
      for (point = 0; point < accum; point++) {
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glTranslatef (xpix * jitter_points[point].x,
		      ypix * jitter_points[point].y, 0.0);

	for (slist = display_slist; slist ; slist = slist->next)
	  {
	    num = *(int*)slist->data;
	    glCallList (num);
	  }
	
	if (g_slist_length (label_display_slist) > 0)
	  {
	    glPopMatrix();		/* labels must not be jittered */
	    glDisable (GL_LIGHTING);
	    
	    for (slist = label_display_slist; slist; slist = slist->next)
	      {
		num = *(int*)slist->data;
		glCallList (num);
	      }
	    glEnable (GL_LIGHTING);
	    glPushMatrix();
	    glTranslatef (xpix * jitter_points[point].x,
			  ypix * jitter_points[point].y, 0.0);
	  }
	
	if (g_slist_length (alpha_display_slist) > 0) 
	  {
	    glEnable (GL_BLEND);
	    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	    
	    for (slist = alpha_display_slist; slist; slist = slist->next)
	      {
		num = *(int*) slist->data;
		glCallList (num);
	      }
	    glDisable (GL_BLEND);
	  }
	
	glPopMatrix();

	if (point == 0) 
	  glAccum (GL_LOAD, fraction);
	else 
	  glAccum (GL_ACCUM, fraction);
	
      }

      glAccum (GL_RETURN, 1.0);

    }
  else
    {

      for (slist = display_slist; slist; slist = slist->next)
	{
	  num = *(int*)slist->data;
	  glCallList (num);
	}
      

      if (g_slist_length (label_display_slist) > 0) 
	{
	  glDisable (GL_LIGHTING);

	  for (slist = label_display_slist; slist; slist = slist->next)
	    {
	      num = *(int*) slist->data;
	      glCallList (num);
	    }
	  
	  glEnable (GL_LIGHTING);
	}

      if (g_slist_length (alpha_display_slist) > 0)
	{
	  glEnable (GL_BLEND);
	  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	  for (slist = alpha_display_slist; slist; slist = slist->next)
	    {
	      num = *(int*)slist->data;
	      glCallList (num);
	    }
	  glDisable (GL_BLEND);
	}
    }
  
}


void
create_dnd_glarea (DNDData *data)
{
  GtkWidget *glarea;
  
  int attr_list[] =
  {
    GDK_GL_RGBA,
    GDK_GL_DOUBLEBUFFER,
    GDK_GL_DEPTH_SIZE,
    GDK_GL_USE_GL,
    GDK_GL_NONE
  };

  if (gdk_gl_query () == FALSE)
    {
      g_print ("OpenGL not supported\n");
      return ;
    }

  glarea = gtk_gl_area_new (attr_list);

  data->glarea = glarea;
  
  gtk_widget_set_events(GTK_WIDGET(glarea),
			GDK_EXPOSURE_MASK|
			GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|
			GDK_POINTER_MOTION_MASK|
			GDK_POINTER_MOTION_HINT_MASK);

  trackball (mesh_info.quat, 0.0, 0.0, 0.0, 0.0);
  glDeleteLists (1, (GLsizei) latest_display_list);

  /* render window */
  gtk_signal_connect (GTK_OBJECT (glarea), "expose_event",
		      GTK_SIGNAL_FUNC (glarea_expose), NULL);
  
  /* motion */
  gtk_signal_connect (GTK_OBJECT (glarea), "motion_notify_event",
		      GTK_SIGNAL_FUNC (glarea_motion_notify), NULL);

  /* mouse action */
  gtk_signal_connect (GTK_OBJECT (glarea), "button_press_event",
		      GTK_SIGNAL_FUNC (glarea_button_press), NULL);

  gtk_signal_connect (GTK_OBJECT (glarea), "button_release_event",
		      GTK_SIGNAL_FUNC (glarea_button_release), NULL);
  
  /* reshape_window */
  gtk_signal_connect (GTK_OBJECT (glarea), "configure_event",
		      GTK_SIGNAL_FUNC (glarea_configure), NULL);
  gtk_signal_connect (GTK_OBJECT (glarea), "destroy",
		      GTK_SIGNAL_FUNC (glarea_destroy), NULL);
  /* Do initialization when widget has been realized. */
  gtk_signal_connect(GTK_OBJECT(glarea), "realize",
		     GTK_SIGNAL_FUNC(init), NULL);

  gtk_widget_set_usize (GTK_WIDGET (glarea), 500, 500);

  gtk_object_set_data (GTK_OBJECT (glarea), "widget", data->window->app);

  data->graphics_info = g_new0 (GraphicsInfo, 1);
  
  dnd_init_graphics_info (data->graphics_info);
  
  gtk_object_set_data (GTK_OBJECT (glarea), "graphics", data->graphics_info);
  gtk_object_set_data (GTK_OBJECT (glarea), "dnd_data", data);
  
}
