#include <GL/gl.h>
#include <GL/glu.h>
#include <math.h>
#include <glib.h>
#include "dnd_main.h"
#include "dnd_glarea.h"
#include "vector3.h"
#include "molecule.h"
#include "ogl_libs.h"

static void ogl_sphere_faces_globe (vector3 *pos, double radius, int segments);
static GLUquadricObj *quadObj;
static void bonds (Molecule *mol, Color **atom_colors, GraphicsInfo *g_info);
static void sticks (Molecule *mol, Color **atom_colors, GraphicsInfo *g_info);
static void sphere (Molecule *mol, Color **atom_colors);
static void set_light_pos (GLfloat *light_pos,
			   GLfloat x, GLfloat y, GLfloat z, GLfloat w);
static void set_light_color (Color *light_color,
			     double r, double g, double b, double a);


static void
set_material_properties ()
{
  GLfloat gl_col[4];
  
  gl_col[0] = 0.;
  gl_col[1] = 1;
  gl_col[2] = 0.;
  gl_col[3] = 1;
  
  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, gl_col);
  glMaterialfv (GL_FRONT_AND_BACK, GL_EMISSION, gl_col);
}

static void
get_atom_colors (DNDData *data, Color **atom_colors)
{
  GList *ls;
  Atom  *atom;
  int    i = 0;
  Color *cc, *nc, *oc, *hc, *sc;
    
  g_return_if_fail (data != NULL);
  g_return_if_fail (atom_colors != NULL);

  cc = named_colors[data->color_info->c_color].color;
  nc = named_colors[data->color_info->n_color].color;
  oc = named_colors[data->color_info->o_color].color;
  hc = named_colors[data->color_info->h_color].color;
  sc = named_colors[data->color_info->s_color].color;
  
  for (ls = data->mol->atom_list; ls; ls = ls->next)
    {
      atom = (Atom *)ls->data;
      
      if (atom->name[0] == 'C')
	atom_colors[i] = cc;
      else if (atom->name[0] == 'N')
	atom_colors[i] = nc;
      else if (atom->name[0] == 'O')
	atom_colors[i] = oc;
      else if (atom->name[0] == 'H')
	atom_colors[i] = hc;
      else if (atom->name[0] == 'S')
	atom_colors[i] = sc;
      else
	{
	  g_error ("Out of Color ... ^^\n");
	  atom_colors[i] = named_colors[10].color;
	}
      i++;
    }
  
}

void
dnd_display_label (Pos *pos, char *label, unsigned int fontbase)
{
  g_return_if_fail (pos != NULL);
  g_return_if_fail (label != NULL);

  start_display_list (TRUE);
  
  glColor3f (1, 1, 0);

  glRasterPos3d (pos->x, pos->y, pos->z);
  glListBase (fontbase);
  glCallLists (strlen (label), GL_UNSIGNED_BYTE, label);
  
  glEndList ();
  
}

void
dnd_display_pick (DNDData *data)
{
  Molecule *mol;
  GraphicsInfo *g_info;

  g_return_if_fail (data != NULL);
  g_return_if_fail (data->mol != NULL);

  mol = data->mol;
  g_info = data->graphics_info;
  
  if (g_info->render_mode == GL_SELECT)
    {
      gint i;
      GList *list;
      Pos   *pos;
      
      start_display_list (FALSE);

      glDisable (GL_LIGHTING);
      glShadeModel (GL_FLAT);

      for (i = 0, list = mol->pos_list; list; i++, list = list->next)
	{
	  pos = (vector3 *)list->data;
      
	  glLoadName (i);
	  glBegin (GL_POINTS);
	  glVertex3d (pos->x, pos->y, pos->z);
	  glEnd();
	}
      
      glShadeModel (GL_SMOOTH);
      glEnable (GL_LIGHTING);
      glEndList ();
    }
  
}

  
void
dnd_display (DNDData *data)
{
  Molecule *mol;
  gint      natom;
  Color   **atom_colors;
  GraphicsInfo *g_info;
  
  g_return_if_fail (data != NULL);
  g_return_if_fail (data->mol != NULL);

  mol = data->mol;
  g_info = data->graphics_info;
  
  natom = g_list_length (mol->atom_list);

  /* --- Set Atom Color list --- */
  atom_colors = g_new0 (Color*, natom);
  get_atom_colors (data, atom_colors);

  switch (g_info->graphics_mode)
    {
    case DND_GRAPHICS_BOND:
      {
	bonds (mol, atom_colors, g_info);
	break;
      }
    case DND_GRAPHICS_SPHERE:
      {
	sphere (mol, atom_colors);
	break;
      }
    case DND_GRAPHICS_CYLINDER:
      {
	sticks (mol, atom_colors, g_info);
	break;
      }
    }
      
}


static void
bonds (Molecule *mol, Color **atom_colors, GraphicsInfo *g_info)
{
  vector3 *pos1, *pos2;
  GList *list;
  Bond *bond;
  
  /* --- Start Display --- */

  start_display_list (FALSE);

  glDisable (GL_LIGHTING);
  glShadeModel (GL_FLAT);

  glLineWidth (g_info->bond_line_width);
  
  glBegin (GL_LINES);

  for (list = mol->bond_list; list; list = list->next)
    {
      vector3 pos;
      gint    n1, n2;
      
      bond = (Bond*)list->data;
        
      n1 = bond->natom[0];
      n2 = bond->natom[1];
      
      pos1 = g_list_nth_data (mol->pos_list, n1);
      pos2 = g_list_nth_data (mol->pos_list, n2);
  
      v3_between (&pos, pos1, pos2, 0.5);

      glBegin (GL_LINES);

      glColor4d (atom_colors[n1]->r, atom_colors[n1]->g,
		 atom_colors[n1]->b, atom_colors[n1]->a);
      glVertex3d (pos1->x, pos1->y, pos1->z);
      glVertex3d (pos.x, pos.y, pos.z);
      
      glColor4d (atom_colors[n2]->r, atom_colors[n2]->g,
		 atom_colors[n2]->b, atom_colors[n2]->a);
      
      glVertex3d (pos.x, pos.y, pos.z);
      glVertex3d (pos2->x, pos2->y, pos2->z);

      glEnd ();

    }
  
  glShadeModel (GL_SMOOTH);
  glEnable (GL_LIGHTING);
  glEndList ();
  
}


static void
sticks (Molecule *mol, Color **atom_colors, GraphicsInfo *g_info)
{
  vector3 *pos1, *pos2;
  GList   *list;
  Bond    *bond;
  
  /* --- Start Display --- */
  start_display_list (FALSE);

  glDisable (GL_LIGHTING);
  glShadeModel (GL_SMOOTH);

  /* [FIXME] */
  /*
  set_material_properties ();
  */

  for (list = mol->bond_list; list; list = list->next)
    {
      vector3 pos;
      gint    n1, n2;

      bond = (Bond*)list->data;
        
      n1 = bond->natom[0];
      n2 = bond->natom[1];
      
      pos1 = g_list_nth_data (mol->pos_list, n1);
      pos2 = g_list_nth_data (mol->pos_list, n2);
  
      v3_between (&pos, pos1, pos2, 0.5);
      
      glColor4d (atom_colors[n1]->r, atom_colors[n1]->g,
		 atom_colors[n1]->b, atom_colors[n1]->a);
      ogl_cylinder_faces (pos1, &pos,
			  g_info->cylinder_radius,
			  g_info->segments,
			  FALSE);
            
      glColor4d (atom_colors[n2]->r, atom_colors[n2]->g,
		 atom_colors[n2]->b, atom_colors[n2]->a);
      ogl_cylinder_faces (&pos, pos2,
			  g_info->cylinder_radius,
			  g_info->segments,
			  FALSE);
    }

  glEnable (GL_LIGHTING);

  glEndList ();
   
}


static void
sphere (Molecule *mol, Color **atom_colors)
{
  int i;
  GList *list;
  vector3   *pos;
  
  start_display_list (FALSE);
  
  glDisable (GL_LIGHTING);
  glShadeModel (GL_SMOOTH);
  /* 
 set_material_properties ();
  */

  for (list = mol->pos_list, i = 0; i < 5; list = list->next, i++)
    {
      pos = (Pos *) list->data;
      
      glColor4d (atom_colors[i]->r, atom_colors[i]->g,
		 atom_colors[i]->b, atom_colors[i]->a);
      ogl_sphere_faces_globe (pos, 0.6, 5);
    }

  glEndList ();

}

void
ogl_sphere_faces_globe (vector3 *pos, double radius, int segments)
     /*
       Render a sphere using the globe subdivision. The segments
       number controls the granularity.
       Requires mode glShadeModel(GL_SMOOTH) for correct rendering.
     */
{
  vector3 point;
  double x, y, z, angle, z1, z2, sr1, sr2, sinangle, cosangle;
  int slices, stacks, slice, stack;

  /* pre */
  g_return_if_fail (pos != NULL);
  g_return_if_fail (radius > 0.0);
  g_return_if_fail (segments >= 2);

  slices = 2 * segments;
  stacks = segments;
  x = pos->x;
  y = pos->y;
  z = pos->z;

  glBegin (GL_TRIANGLE_FAN);
  point.x = 0.0;
  point.y = 0.0;
  point.z = 1.0;
  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);
  angle = M_PI / (double) stacks;
  sr1 = sin (angle);
  z1 = cos (angle);
  point.x = sr1;
  point.y = 0.0;
  point.z = z1;
  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);

  for (slice = 1; slice < slices; slice++) {
    angle = (double) slice * 2.0 * M_PI / (double) slices;
    point.x = sr1 * cos (angle);
    point.y = sr1 * sin (angle);
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);
  }
  point.x = sr1;
  point.y = 0.0;
  point.z = z1;

  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);
  glEnd();

  for (stack = 1; stack < stacks - 1; stack++) {
    glBegin (GL_QUAD_STRIP);
    angle = (double) (stack + 1) * M_PI / (double) stacks;
    sr2 = sin (angle);
    z2 = cos (angle);
    point.x = sr2;
    point.y = 0.0;
    point.z = z2;
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);
   
    angle = (double) stack * M_PI / (double) stacks;
    sr1 = sin (angle);
    z1 = cos (angle);
    point.x = sr1;
    point.y = 0.0;
    point.z = z1;
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);
   
    for (slice = slices - 1; slice >= 1; slice--) {
      angle = (double) slice * 2.0 * M_PI / (double) slices;
      cosangle = cos (angle);
      sinangle = sin (angle);
      point.x = sr2 * cosangle;
      point.y = sr2 * sinangle;
      point.z = z2;
      glNormal3d (point.x, point.y, point.z);
      glVertex3d (radius * point.x, radius * point.y, radius *point.z);
   
      point.x = sr1 * cosangle;
      point.y = sr1 * sinangle;
      point.z = z1;
      glNormal3d (point.x, point.y, point.z);
      glVertex3d (radius * point.x, radius * point.y, radius *point.z);
    }
    point.x = sr2;
    point.y = 0.0;
    point.z = z2;
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);

    point.x = sr1;
    point.y = 0.0;
    point.z = z1;
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);

    glEnd();
  }

  glBegin (GL_TRIANGLE_FAN);
  point.x = 0.0;
  point.y = 0.0;
  point.z = -1.0;
  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);

  angle = M_PI / (double) stacks;
  sr1 = sin (angle);
  z1 = - cos (angle);
  point.x = sr1;
  point.y = 0.0;
  point.z = z1;
  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);

  for (slice = slices - 1; slice >= 1; slice--) {
    angle = (double) slice * 2.0 * M_PI / (double) slices;
    point.x = sr1 * cos (angle);
    point.y = sr1 * sin (angle);
    glNormal3d (point.x, point.y, point.z);
    glVertex3d (radius * point.x, radius * point.y, radius *point.z);
  }
  point.x = sr1;
  point.y = 0.0;
  point.z = z1;
  glNormal3d (point.x, point.y, point.z);
  glVertex3d (radius * point.x, radius * point.y, radius *point.z);
  glEnd();
}


void
dnd_init_graphics_info (GraphicsInfo *g_info)
{
  LightInfo *light_info;
  
  g_return_if_fail (g_info != NULL);
  
  g_info->graphics_mode = DND_GRAPHICS_BOND;
  g_info->render_mode   = GL_RENDER;
  
  g_info->bond_line_width = 1.0;
  
  g_info->coil_radius = 1.5;
  
  g_info->cylinder_radius = 0.2;
  
  g_info->segments  = 10;
  
  light_info = g_new0 (LightInfo, 1);
  light_info->light_enum = GL_LIGHT0;
  set_light_pos (light_info->light_pos, +50, 0, 0, 0);
  set_light_color (&light_info->light_color, .6, .6, .6, 1.0);
  g_info->light_list = g_list_append (g_info->light_list, light_info);

  light_info = g_new0 (LightInfo, 1);
  light_info->light_enum = GL_LIGHT1;
  set_light_pos (light_info->light_pos, -50, 0, 0, 0);
  set_light_color (&light_info->light_color, .4, .4, 1, 1.0);
  g_info->light_list = g_list_append (g_info->light_list, light_info);
}



static void
set_light_pos (GLfloat *light_pos, GLfloat x, GLfloat y, GLfloat z, GLfloat w)
{
  light_pos[0] = x;
  light_pos[1] = y;
  light_pos[2] = z;
  light_pos[3] = w;
}


static void
set_light_color (Color *light_color,
		 double r, double g, double b, double a)
{
  light_color->r = r;
  light_color->g = g;
  light_color->b = b;
  light_color->a = a;
}
  

