#include "cell.h"


gint 
cell_bin (double rc, gint nc, double esp)
{
  int ibin;
  
  if (rc < -0.5+eps || rc >= 0.5-eps)
    {
      if (rc < -0.5+eps && rc >= -0.5-eps)
	rc = -0.5;
      else if(rc >= 0.5-eps && rc <= 0.5+eps)
	rc = 0.5-eps;
      else 
	g_error ("Co-ordinate out of range in BIN (fill_cells) %.17g\n",rc);
    }

  ibin = (rc+0.5)*nc;
  
  if(ibin >= nc || ibin < 0)
    g_error ("Rounding problem in BIN (fill_cells) %.17g\n",rc);

  return ibin;
}

 
gint
cell_locate (Pos *pos, int cell_nx, int cell_ny, int cell_nz, double esp)
{
  gint ix, iy, iz;
  gint icell;
  gint nx, ny, nz;
  
  ix = cell_bin (pos->x, cell_nx, esp);
  iy = cell_bin (pos->y, cell_ny, esp);
  iz = cell_bin (pos->z, cell_nz, esp);
  
  icell = iz + (cell_nz) * (iy + (cell_ny) * ix);
  
  return icell;
}


void
cell_free (Cell **gcell)
{
  Cell *cell;
  
  g_return_if_fail (gcell != NULL);
  
  cell = *gcell;
  
  /* [FIXME] */
  g_list_free (cell->linked_list);
  g_list_free (cell->neighbor_list);
  
  g_free (cell);
}


  
static void
cell_update (System *system)
{
  GList *list;
  Cell  *cell;
  int    ncells;
  int    i;
  
  /* free cell_list */
  if (system->cell_list != NULL)
    {
      GList *tmp_list;
      
      list = system->cell_list;
      
      while (list)
	{
	  tmp_list = list;
	  list = list->next;
	  
	  cell = tmp_list->data;
	  cell_free (&cell);
	  
	  tmp_list->data = NULL;
	}

      g_list_free (system->cell_list);
      system->cell_list = NULL;
    }

  /* allocate cell_list */

  list = system->cell_list;
      
  for (i = 0; i < ncells; i++)
    {
      cell = g_new0 (Cell, 1);
      
      list = g_list_append (list, cell);
      
      if (system->cell_list == NULL)
	system->cell_list = list;
      else
	list = list->next;
      
    }

  /* -- system->cell_pair_list --- */

  g_list_free (system->cell_pair_list);
  system->cell_pair_list = NULL;
  
}



void
cell_init (MDData *data)
{
  static double subcell = 0.0;
  static int onx, ony, onz;
  int        nx,  ny,  nz;
  int        ncells;
  GList      *list;
  
  g_return_if_fail (control != NULL);
  g_return_if_fail (system != NULL);
  
  if (subcell <= 0)
    subcell = control->cutoff / 5.0;
  
  nx = system->h0.x / subcell + 0.5;
  ny = system->h1.y / subcell + 0.5;
  nz = system->h2.z / subcell + 0.5;

  ncells = nx * ny * nz;
  
  if (nx != system->cell_nx ||
      ny != system->cell_ny ||
      nz != system->cell_nz)
    {
      g_message ("MD cell divided into %d subcells (%dx%dx%d)",
		 ncells, nx, ny, nz);

      system->cell_nx = nx;
      system->cell_ny = ny;
      system->cell_nz = nz;

      system->ncells = ncells;
      
      cell_update (data->system);
    }

  cell_setup_linked_list (data->system, data->mol);
  cell_setup_neighbor_list (data->system, data->mol);
  
}


static void
cell_setup_linked_list (System *system, Molecule *mol)
{
  gint ncells;
  gint nx, ny, nz;
  Cell *cell;
  
  ncells = system->ncells;
  nx     = system->cell_nx;
  ny     = system->cell_ny;
  nz     = system->cell_nz;
  
  for (plist = mol->pos_list, alist = mol->atom_list;
       plist;
       plist = plist->next, alist = alist->next)
    {
      pos = plist->data;
      atom = rlist->data;
      
      icell = cell_locate (pos, nx, ny, nz, eps);
      cell = g_list_nth_data (system->cell_list, icell);
      
      cell_linked = g_new0 (CellLinked, 1);
      
      cell_linked->atom = atom;
      cell_linked->pos  = pos;
      
      cell->linked_list = g_list_append (cell->linked_list, cell_linked);
      
    }
  
}


void
cell_setup_neighbor_list (System *system)
{
  int nx, ny, nz;
  int icell;
  GList *list;
  
  nx = system->cell_nx;
  ny = system->cell_ny;
  nz = system->cell_nz;

  /* system->cell_pair_list = NULL */
  list = system->cell_pair_list;
    
  for (iz = 0; iz < nz; iz++)
    for (iy = 0; iy < ny; iy++)
      for (ix = 0; ix < nx; ix++)
	{
	  list = cell_fill_neighbor_list (system->cell_list, list,
					  ix, iy, iz, nx, ny, nz);
	}
  
}

/*
  return cell_pair_list
*/

static GList *
cell_fill_pair_list (GList *cell_list, GList *cell_pair_list,
		     gint ix, gint iy, gint iz,
		     gint nx, gint ny, gint nz)
{
#define ICELL(ix, iy, iz) (iz + nz * (iy + ny * ix))
  gint icell;
  gint dc;   /* cell number between vdw radius */
  GList *list;
  Cell  *cell1, *cell2;
  CellPair *cell_pair;
  gboolean is_mirror;
  
  dc = 2;
  icell = ICELL (ix, iy, iz);
  cell1 = g_list_nth_data (cell_list, icell);
  
  for (x = ix; x < ix+dc; x++)
    for (y = iy; y < iy+dc; y++)
      for (z = iz; z < iz+dc; z++)
	{
	  if (x == ix && y == iy && z == iz)
	    continue;
	  
	  cell_pair = g_new0 (CellPair, 1);
	  
	  is_mirror = FALSE;
	  
	  if (x+1 == nx)
	    { 
	      x = 0;
	      is_mirror = TRUE;
	    }
	  
	  if (y+1 == ny)
	    {
	      y = 0;
	      is_mirror = TRUE;
	    }
	  
	  if (z+1 == nz)
	    {
	      z = 0;
	      is_mirror = TRUE;
	    }
	  
	  icell = ICELL (x, y, z);
	  cell2 = g_list_nth_data (cell_list, icell);
	  
	  cell_pair->cell1 = cell1;
	  cell_pair->cell2 = cell2;
	  cell_pair->is_mirror = is_mirror;
	  
	  cell_pair_list = g_list_prepend (cell_pair_list, cell_pair);
	}

  return cell_pair_list;
  
}

  
static guint
cell_hash_func (gconstpointer key)
{
  return (guint) *((const gint*) key);
}

static guint
cell_hash_compare (gconstpointer a,
		   gconstpointer b)
{
  return *((const gint*) a) == *((const gint*) b);
}
